# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.pynumero.interfaces.nlp import NLP
from pyomo.contrib.pynumero.sparse import BlockVector, BlockMatrix
from pyomo.contrib.pynumero.linalg.ma27_interface import MA27
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
import numpy as np
from scipy.sparse import tril
import pyomo.environ as pyo
from pyomo import dae
from pyomo.common.timing import TicTocTimer
import time
from pyomo.contrib.pynumero.linalg.base import LinearSolverInterface, LinearSolverStatus


def build_burgers_model(nfe_x=100, nfe_t=200, start_t=0, end_t=1):
    """
    Build a pyomo.dae model of an optimal control problem with Burgers' equation

    Parameters
    ----------
    nfe_x: int
        Number of finite elements for discretization of x
    nfe_t: int
        Number of finite elements for discretization of t
    start_t: float
        The start of the time horizon
    end_t: float
        The end of the time horizon

    Returns
    -------
    m: Pyomo model
    """
    dt = (end_t - start_t) / float(nfe_t)

    start_x = 0
    end_x = 1
    dx = (end_x - start_x) / float(nfe_x)

    m = pyo.Block(concrete=True)
    m.omega = pyo.Param(initialize=0.02)
    m.v = pyo.Param(initialize=0.01)
    m.r = pyo.Param(initialize=0)

    m.x = dae.ContinuousSet(bounds=(start_x, end_x))
    m.t = dae.ContinuousSet(bounds=(start_t, end_t))

    m.y = pyo.Var(m.x, m.t)
    m.dydt = dae.DerivativeVar(m.y, wrt=m.t)
    m.dydx = dae.DerivativeVar(m.y, wrt=m.x)
    m.dydx2 = dae.DerivativeVar(m.y, wrt=(m.x, m.x))

    m.u = pyo.Var(m.x, m.t)

    def _y_init_rule(m, x):
        if x <= 0.5 * end_x:
            return 1
        return 0

    m.y0 = pyo.Param(m.x, default=_y_init_rule)

    def _upper_x_bound(m, t):
        return m.y[end_x, t] == 0

    m.upper_x_bound = pyo.Constraint(m.t, rule=_upper_x_bound)

    def _lower_x_bound(m, t):
        return m.y[start_x, t] == 0

    m.lower_x_bound = pyo.Constraint(m.t, rule=_lower_x_bound)

    def _upper_x_ubound(m, t):
        return m.u[end_x, t] == 0

    m.upper_x_ubound = pyo.Constraint(m.t, rule=_upper_x_ubound)

    def _lower_x_ubound(m, t):
        return m.u[start_x, t] == 0

    m.lower_x_ubound = pyo.Constraint(m.t, rule=_lower_x_ubound)

    def _lower_t_bound(m, x):
        if x == start_x or x == end_x:
            return pyo.Constraint.Skip
        return m.y[x, start_t] == m.y0[x]

    def _lower_t_ubound(m, x):
        if x == start_x or x == end_x:
            return pyo.Constraint.Skip
        return m.u[x, start_t] == 0

    m.lower_t_bound = pyo.Constraint(m.x, rule=_lower_t_bound)
    m.lower_t_ubound = pyo.Constraint(m.x, rule=_lower_t_ubound)

    # PDE
    def _pde(m, x, t):
        if t == start_t or x == end_x or x == start_x:
            e = pyo.Constraint.Skip
        else:
            # print(foo.last_t, t-dt, abs(foo.last_t - (t-dt)))
            # assert math.isclose(foo.last_t, t - dt, abs_tol=1e-6)
            e = (
                m.dydt[x, t] - m.v * m.dydx2[x, t] + m.dydx[x, t] * m.y[x, t]
                == m.r + m.u[x, t]
            )
        return e

    m.pde = pyo.Constraint(m.x, m.t, rule=_pde)

    # Discretize Model
    disc = pyo.TransformationFactory('dae.finite_difference')
    disc.apply_to(m, nfe=nfe_t, wrt=m.t, scheme='BACKWARD')
    disc.apply_to(m, nfe=nfe_x, wrt=m.x, scheme='CENTRAL')

    # Solve control problem using Pyomo.DAE Integrals
    def _intX(m, x, t):
        return (m.y[x, t] - m.y0[x]) ** 2 + m.omega * m.u[x, t] ** 2

    m.intX = dae.Integral(m.x, m.t, wrt=m.x, rule=_intX)

    def _intT(m, t):
        return m.intX[t]

    m.intT = dae.Integral(m.t, wrt=m.t, rule=_intT)

    def _obj(m):
        e = 0.5 * m.intT
        for x in sorted(m.x):
            if x == start_x or x == end_x:
                pass
            else:
                e += 0.5 * 0.5 * dx * dt * m.omega * m.u[x, start_t] ** 2
        return e

    m.obj = pyo.Objective(rule=_obj)

    return m


def sqp(
    nlp: NLP, linear_solver: LinearSolverInterface, max_iter=100, tol=1e-8, output=True
):
    """
    An example of a simple SQP algorithm for
    equality-constrained NLPs.

    Parameters
    ----------
    nlp: NLP
        A PyNumero NLP
    max_iter: int
        The maximum number of iterations
    tol: float
        The convergence tolerance
    """
    t0 = time.time()

    # setup KKT matrix
    kkt = BlockMatrix(2, 2)
    rhs = BlockVector(2)

    # create and initialize the iteration vector
    z = BlockVector(2)
    z.set_block(0, nlp.get_primals())
    z.set_block(1, nlp.get_duals())

    if output:
        print(
            f"{'Iter':<12}{'Objective':<12}{'Primal Infeasibility':<25}{'Dual Infeasibility':<25}{'Elapsed Time':<15}"
        )

    # main iteration loop
    for _iter in range(max_iter):
        nlp.set_primals(z.get_block(0))
        nlp.set_duals(z.get_block(1))

        grad_lag = (
            nlp.evaluate_grad_objective()
            + nlp.evaluate_jacobian_eq().transpose() * z.get_block(1)
        )
        residuals = nlp.evaluate_eq_constraints()

        if output:
            print(
                f"{_iter:<12}{nlp.evaluate_objective():<12.2e}{np.abs(residuals).max():<25.2e}{np.abs(grad_lag).max():<25.2e}{time.time()-t0:<15.2e}"
            )

        if np.abs(grad_lag).max() <= tol and np.abs(residuals).max() <= tol:
            break

        kkt.set_block(0, 0, nlp.evaluate_hessian_lag())
        kkt.set_block(1, 0, nlp.evaluate_jacobian_eq())
        kkt.set_block(0, 1, nlp.evaluate_jacobian_eq().transpose())

        rhs.set_block(0, grad_lag)
        rhs.set_block(1, residuals)

        delta, res = linear_solver.solve(kkt, -rhs)
        assert res.status == LinearSolverStatus.successful
        z += delta


def load_solution(m: pyo.ConcreteModel(), nlp: PyomoNLP):
    primals = nlp.get_primals()
    pyomo_vars = nlp.get_pyomo_variables()
    for v, val in zip(pyomo_vars, primals):
        v.value = val


def main(linear_solver, nfe_x=100, nfe_t=200):
    m = build_burgers_model(nfe_x=nfe_x, nfe_t=nfe_t)
    nlp = PyomoNLP(m)
    sqp(nlp, linear_solver)
    load_solution(m, nlp)
    return pyo.value(m.obj)


if __name__ == '__main__':
    # create the linear solver
    linear_solver = MA27()
    linear_solver.set_cntl(1, 1e-6)  # pivot tolerance

    optimal_obj = main(linear_solver)
    print(f'Optimal Objective: {optimal_obj}')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base import Transformation, TransformationFactory
import pyomo.gdp.plugins.multiple_bigm


@TransformationFactory.register(
    'contrib.piecewise.convex_combination',
    doc="Convert piecewise-linear model to a GDP "
    "to 'Convex Combination' MIP formulation.",
)
class ConvexCombinationTransformation(Transformation):
    """
    Converts a model containing PiecewiseLinearFunctions to a an equivalent
    MIP via the Convex Combination method from [1]. Note that,
    while this model probably resolves to the model described in [1] after
    presolve, the Pyomo version is not as simplified.

    References
    ----------
    [1] J.P. Vielma, S. Ahmed, and G. Nemhauser, "Mixed-integer models
        for nonseparable piecewise-linear optimization: unifying framework
        and extensions," Operations Research, vol. 58, no. 2, pp. 305-315,
        2010.
    """

    def _apply_to(self, instance, **kwds):
        TransformationFactory('contrib.piecewise.reduced_inner_repn_gdp').apply_to(
            instance
        )
        TransformationFactory('gdp.mbigm').apply_to(
            instance, reduce_bound_constraints=True
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Param,
    Var,
    Objective,
    Constraint,
    NonNegativeReals,
    SOSConstraint,
    sum_product,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class SOS1_simple(_BaseTestModel):
    """
    A simple linear program
    """

    description = "SOS1_simple"
    level = ('nightly', 'expensive')
    capabilities = set(['linear', 'integer', 'sos1'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.a = Param(initialize=0.1)
        model.x = Var(within=NonNegativeReals)
        model.y = Var([1, 2], within=NonNegativeReals)

        model.obj = Objective(expr=model.x + model.y[1] + 2 * model.y[2])
        model.c1 = Constraint(expr=model.a <= model.y[2])
        model.c2 = Constraint(expr=(2.0, model.x, 10.0))
        model.c3 = SOSConstraint(var=model.y, index=[1, 2], sos=1)
        model.c4 = Constraint(expr=sum_product(model.y) == 1)

        # Make an empty SOSConstraint
        model.c5 = SOSConstraint(var=model.y, index=[1, 2], sos=1)
        model.c5.set_items([], [])
        assert len(list(model.c5.get_items())) == 0

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = 0
        model.y[1].value = 1
        model.y[2].value = None


@register_model
class SOS1_simple_kernel(SOS1_simple):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.a = pmo.parameter(value=0.1)
        model.x = pmo.variable(domain=NonNegativeReals)
        model.y = pmo.variable_dict()
        model.y[1] = pmo.variable(domain=NonNegativeReals)
        model.y[2] = pmo.variable(domain=NonNegativeReals)

        model.obj = pmo.objective(model.x + model.y[1] + 2 * model.y[2])
        model.c1 = pmo.constraint(model.a <= model.y[2])
        model.c2 = pmo.constraint((2.0, model.x, 10.0))
        model.c3 = pmo.sos1(model.y.values())
        model.c4 = pmo.constraint(sum(model.y.values()) == 1)

        # Make an empty SOS constraint
        model.c5 = pmo.sos1([])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from pyomo.common.dependencies import numpy as np

from pyomo.contrib.doe.examples.reactor_experiment import ReactorExperiment
from pyomo.contrib.doe import DesignOfExperiments
from pyomo.contrib.doe import utils

from pyomo.contrib.parmest.utils.model_utils import update_model_from_suffix
from os.path import join, abspath, dirname

import pyomo.environ as pyo

import json


# Example to run a DoE on the reactor
def main():
    # Read in file
    file_dirname = dirname(abspath(str(__file__)))
    file_path = abspath(join(file_dirname, "result.json"))

    # Read in data
    with open(file_path) as f:
        data_ex = json.load(f)

    # Put temperature control time points into correct format for reactor experiment
    data_ex["control_points"] = {
        float(k): v for k, v in data_ex["control_points"].items()
    }

    # Create a ReactorExperiment object; data and discretization information are part
    # of the constructor of this object
    experiment = ReactorExperiment(data=data_ex, nfe=10, ncp=3)

    # Call the experiment's model using get_labeled_model
    reactor_model = experiment.get_labeled_model()

    # Show the model
    reactor_model.pprint()
    # The suffix object 'measurement_error' stores measurement error values for each component.
    # Here, we retrieve the original values from the suffix for inspection.
    suffix_obj = reactor_model.measurement_error
    me_vars = list(suffix_obj.keys())  # components
    orig_vals = np.array(list(suffix_obj.values()))

    # Original values
    print("Original sigma values")
    print("-----------------------")
    suffix_obj.display()

    # Update the suffix with new values
    new_vals = orig_vals + 1
    # Here we are updating the values of the measurement error
    # You must know the length of the list and order of the suffix items to update them correctly
    update_model_from_suffix(suffix_obj, new_vals)

    # Updated values
    print("Updated sigma values :")
    print("-----------------------")
    suffix_obj.display()
    return suffix_obj, orig_vals, new_vals


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from typing import List, Mapping

import pyomo.environ as pyo
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.objective import ObjectiveData
from pyomo.core.base.param import ParamData
from pyomo.core.base.sos import SOSConstraintData
from pyomo.core.base.var import VarData
from pyomo.common import unittest
from pyomo.contrib.observer.model_observer import (
    Observer,
    ModelChangeDetector,
    AutoUpdateConfig,
    Reason,
)
from pyomo.common.collections import DefaultComponentMap, ComponentMap
from pyomo.common.errors import PyomoException

logger = logging.getLogger(__name__)


def make_count_dict():
    d = {i: 0 for i in Reason}
    return d


class ObserverChecker(Observer):
    def __init__(self):
        super().__init__()
        self.counts = DefaultComponentMap(make_count_dict)
        """
        counts is a mapping from component (e.g., variable) to another 
        mapping from Reason to an int that indicates the number of times 
        the corresponding method has been called
        """

    def check(self, expected):
        unittest.assertStructuredAlmostEqual(
            first=expected, second=self.counts, places=7
        )

    def pprint(self):
        for k, d in self.counts.items():
            print(f'{k}:')
            for a, v in d.items():
                print(f'  {a}: {v}')

    def _update_variables(self, variables: Mapping[VarData, Reason]):
        for v, reason in variables.items():
            self.counts[v][reason] += 1

    def _update_parameters(self, params: Mapping[ParamData, Reason]):
        for p, reason in params.items():
            self.counts[p][reason] += 1

    def _update_constraints(self, cons: Mapping[ConstraintData, Reason]):
        for c, reason in cons.items():
            self.counts[c][reason] += 1

    def _update_sos_constraints(self, cons: Mapping[SOSConstraintData, Reason]):
        for c, reason in cons.items():
            self.counts[c][reason] += 1

    def _update_objectives(self, objs: Mapping[ObjectiveData, Reason]):
        for obj, reason in objs.items():
            self.counts[obj][reason] += 1


class TestChangeDetector(unittest.TestCase):
    def test_objective(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.p = pyo.Param(mutable=True, initialize=1)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])

        expected = DefaultComponentMap(make_count_dict)
        obs.check(expected)

        m.obj = pyo.Objective(expr=m.x**2 + m.p * m.y**2)
        detector.update()
        expected[m.obj][Reason.added] += 1
        expected[m.x][Reason.added] += 1
        expected[m.y][Reason.added] += 1
        expected[m.p][Reason.added] += 1
        obs.check(expected)

        m.y.setlb(0)
        detector.update()
        expected[m.y][Reason.bounds] += 1
        obs.check(expected)

        m.x.fix(2)
        detector.update()
        expected[m.x][Reason.fixed] += 1
        obs.check(expected)

        m.x.unfix()
        detector.update()
        expected[m.x][Reason.fixed] += 1
        obs.check(expected)

        m.p.value = 2
        detector.update()
        expected[m.p][Reason.value] += 1
        obs.check(expected)

        m.obj.expr = m.x**2 + m.y**2
        detector.update()
        expected[m.obj][Reason.expr] += 1
        expected[m.p][Reason.removed] += 1
        obs.check(expected)

        expected[m.obj][Reason.removed] += 1
        del m.obj
        m.obj2 = pyo.Objective(expr=m.p * m.x)
        detector.update()
        # remember, m.obj is a different object now
        expected[m.obj2][Reason.added] += 1
        expected[m.y][Reason.removed] += 1
        expected[m.p][Reason.added] += 1
        obs.check(expected)

    def test_constraints(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.p = pyo.Param(mutable=True, initialize=1)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])

        expected = DefaultComponentMap(make_count_dict)
        obs.check(expected)

        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= (m.x - m.p) ** 2)
        detector.update()
        expected[m.x][Reason.added] += 1
        expected[m.y][Reason.added] += 1
        expected[m.p][Reason.added] += 1
        expected[m.c1][Reason.added] += 1
        expected[m.obj][Reason.added] += 1
        obs.check(expected)

        m.x.fix(1)
        detector.update()
        expected[m.x][Reason.fixed] += 1
        obs.check(expected)

        m.z = pyo.Var()
        m.c1.set_value(m.y == 2 * m.z)
        detector.update()
        expected[m.z][Reason.added] += 1
        expected[m.c1][Reason.expr] += 1
        expected[m.p][Reason.removed] += 1
        expected[m.x][Reason.removed] += 1
        obs.check(expected)

        expected[m.c1][Reason.removed] += 1
        del m.c1
        detector.update()
        expected[m.z][Reason.removed] += 1
        obs.check(expected)

    def test_sos(self):
        m = pyo.ConcreteModel()
        m.a = pyo.Set(initialize=[1, 2, 3], ordered=True)
        m.x = pyo.Var(m.a, within=pyo.Binary)
        m.y = pyo.Var(within=pyo.Binary)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.SOSConstraint(var=m.x, sos=1)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])

        expected = DefaultComponentMap(make_count_dict)
        expected[m.obj][Reason.added] += 1
        for i in m.a:
            expected[m.x[i]][Reason.added] += 1
        expected[m.y][Reason.added] += 1
        expected[m.c1][Reason.added] += 1
        obs.check(expected)

        detector.update()
        obs.check(expected)

        m.c1.set_items([m.x[2], m.x[1], m.x[3]], [1, 2, 3])
        detector.update()
        expected[m.c1][Reason.sos_items] += 1
        obs.check(expected)

        m.c1.set_items([m.x[2], m.x[1]], [1, 2])
        detector.update()
        expected[m.c1][Reason.sos_items] += 1
        expected[m.x[3]][Reason.removed] += 1
        obs.check(expected)

        m.c1.set_items([m.x[2], m.x[1], m.x[3]], [1, 2, 3])
        detector.update()
        expected[m.c1][Reason.sos_items] += 1
        expected[m.x[3]][Reason.added] += 1
        obs.check(expected)

        for i in m.a:
            expected[m.x[i]][Reason.removed] += 1
        expected[m.c1][Reason.removed] += 1
        del m.c1
        detector.update()
        obs.check(expected)

    def test_vars_and_params_elsewhere(self):
        m1 = pyo.ConcreteModel()
        m1.x = pyo.Var()
        m1.y = pyo.Var()
        m1.p = pyo.Param(mutable=True, initialize=1)

        m2 = pyo.ConcreteModel()

        obs = ObserverChecker()
        detector = ModelChangeDetector(m2, [obs])

        expected = DefaultComponentMap(make_count_dict)
        obs.check(expected)

        m2.obj = pyo.Objective(expr=m1.y)
        m2.c1 = pyo.Constraint(expr=m1.y >= (m1.x - m1.p) ** 2)
        detector.update()
        expected[m1.x][Reason.added] += 1
        expected[m1.y][Reason.added] += 1
        expected[m1.p][Reason.added] += 1
        expected[m2.c1][Reason.added] += 1
        expected[m2.obj][Reason.added] += 1
        obs.check(expected)

        m1.x.fix(1)
        detector.update()
        expected[m1.x][Reason.fixed] += 1
        obs.check(expected)

    def test_named_expression(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.p = pyo.Param(mutable=True, initialize=1)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])

        expected = DefaultComponentMap(make_count_dict)
        obs.check(expected)

        m.obj = pyo.Objective(expr=m.y)
        m.e = pyo.Expression(expr=m.x - m.p)
        m.c1 = pyo.Constraint(expr=m.y >= m.e)
        detector.update()
        expected[m.x][Reason.added] += 1
        expected[m.y][Reason.added] += 1
        expected[m.p][Reason.added] += 1
        expected[m.c1][Reason.added] += 1
        expected[m.obj][Reason.added] += 1
        obs.check(expected)

        # now modify the named expression and make sure the
        # constraint gets removed and added
        m.e.expr = (m.x - m.p) ** 2
        detector.update()
        expected[m.c1][Reason.expr] += 1
        obs.check(expected)

    def test_update_config(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.p = pyo.Param(initialize=1, mutable=True)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])
        expected = DefaultComponentMap(make_count_dict)
        obs.check(expected)

        detector.config.check_for_new_or_removed_constraints = False
        detector.config.check_for_new_or_removed_objectives = False
        detector.config.update_constraints = False
        detector.config.update_objectives = False
        detector.config.update_vars = False
        detector.config.update_parameters = False
        detector.config.update_named_expressions = False

        m.e = pyo.Expression(expr=pyo.exp(m.x))
        m.obj = pyo.Objective(expr=m.x**2 + m.p * m.y**2)
        m.c1 = pyo.Constraint(expr=m.y >= m.e + m.p)

        detector.update()
        obs.check(expected)

        detector.config.check_for_new_or_removed_constraints = True
        detector.update()
        expected[m.x][Reason.added] += 1
        expected[m.y][Reason.added] += 1
        expected[m.p][Reason.added] += 1
        expected[m.c1][Reason.added] += 1
        obs.check(expected)

        detector.config.check_for_new_or_removed_objectives = True
        detector.update()
        expected[m.obj][Reason.added] += 1
        obs.check(expected)

        m.x.setlb(0)
        detector.update()
        obs.check(expected)

        detector.config.update_vars = True
        detector.update()
        expected[m.x][Reason.bounds] += 1
        obs.check(expected)

        m.p.value = 2
        detector.update()
        obs.check(expected)

        detector.config.update_parameters = True
        detector.update()
        expected[m.p][Reason.value] += 1
        obs.check(expected)

        m.e.expr += 1
        detector.update()
        obs.check(expected)

        detector.config.update_named_expressions = True
        detector.update()
        expected[m.c1][Reason.expr] += 1
        obs.check(expected)

        m.obj.expr += 1
        detector.update()
        obs.check(expected)

        detector.config.update_objectives = True
        detector.update()
        expected[m.obj][Reason.expr] += 1
        obs.check(expected)

        m.c1 = m.y >= m.e
        detector.update()
        obs.check(expected)

        detector.config.update_constraints = True
        detector.update()
        expected[m.c1][Reason.expr] += 1
        obs.check(expected)

    def test_param_in_bounds(self):
        m = pyo.ConcreteModel()
        m.y = pyo.Var()
        m.p = pyo.Param(mutable=True, initialize=1)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])

        expected = DefaultComponentMap(make_count_dict)
        obs.check(expected)

        m.obj = pyo.Objective(expr=m.y)
        m.y.setlb(m.p - 1)
        detector.update()
        expected[m.y][Reason.added] += 1
        expected[m.p][Reason.added] += 1
        expected[m.obj][Reason.added] += 1
        obs.check(expected)

        m.p.value = 2
        detector.update()
        expected[m.p][Reason.value] += 1
        obs.check(expected)

        m.p2 = pyo.Param(mutable=True, initialize=1)
        m.y.setub(m.p2 + 1)
        detector.update()
        expected[m.p2][Reason.added] += 1
        expected[m.y][Reason.bounds] += 1
        obs.check(expected)

    def test_incidence(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.p1 = pyo.Param(mutable=True, initialize=1)
        m.p2 = pyo.Param(mutable=True, initialize=1)
        m.x.setlb(m.p1)

        m.e1 = pyo.Expression(expr=m.x + m.p1)
        m.e2 = pyo.Expression(expr=(m.e1**2))
        m.obj = pyo.Objective(expr=m.e2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.z + m.p2 == 0)
        m.c2 = pyo.Constraint(expr=m.x + m.p2 == 0)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])

        expected = DefaultComponentMap(make_count_dict)
        expected[m.x][Reason.added] += 1
        expected[m.y][Reason.added] += 1
        expected[m.z][Reason.added] += 1
        expected[m.p1][Reason.added] += 1
        expected[m.p2][Reason.added] += 1
        expected[m.obj][Reason.added] += 1
        expected[m.c1][Reason.added] += 1
        expected[m.c2][Reason.added] += 1
        obs.check(expected)

        self.assertEqual(detector.get_variables_impacted_by_param(m.p1), [m.x])
        self.assertEqual(detector.get_variables_impacted_by_param(m.p2), [])
        self.assertEqual(detector.get_constraints_impacted_by_param(m.p1), [])
        self.assertEqual(detector.get_constraints_impacted_by_param(m.p2), [m.c1, m.c2])
        self.assertEqual(detector.get_constraints_impacted_by_var(m.x), [m.c2])
        self.assertEqual(detector.get_constraints_impacted_by_var(m.y), [])
        self.assertEqual(detector.get_constraints_impacted_by_var(m.z), [m.c1])
        self.assertEqual(detector.get_objectives_impacted_by_param(m.p1), [m.obj])
        self.assertEqual(detector.get_objectives_impacted_by_param(m.p2), [])
        self.assertEqual(detector.get_objectives_impacted_by_var(m.x), [m.obj])
        self.assertEqual(detector.get_objectives_impacted_by_var(m.y), [m.obj])
        self.assertEqual(detector.get_objectives_impacted_by_var(m.z), [])

        m.e1.expr += m.z
        detector.update()
        expected[m.obj][Reason.expr] += 1
        obs.check(expected)

        self.assertEqual(detector.get_objectives_impacted_by_param(m.p1), [m.obj])
        self.assertEqual(detector.get_objectives_impacted_by_param(m.p2), [])
        self.assertEqual(detector.get_objectives_impacted_by_var(m.x), [m.obj])
        self.assertEqual(detector.get_objectives_impacted_by_var(m.y), [m.obj])
        self.assertEqual(detector.get_objectives_impacted_by_var(m.z), [m.obj])

    def test_manual_updates(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.p = pyo.Param(mutable=True, initialize=1)

        obs = ObserverChecker()
        detector = ModelChangeDetector(m, [obs])

        expected = DefaultComponentMap(make_count_dict)
        obs.check(expected)

        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= (m.x - m.p) ** 2)
        m.c2 = pyo.Constraint(expr=m.x + m.y == 0)

        detector.add_objectives([m.obj])
        expected[m.obj][Reason.added] += 1
        expected[m.y][Reason.added] += 1
        obs.check(expected)

        detector.add_constraints([m.c1])
        expected[m.x][Reason.added] += 1
        expected[m.p][Reason.added] += 1
        expected[m.c1][Reason.added] += 1
        obs.check(expected)

        detector.add_constraints([m.c2])
        expected[m.c2][Reason.added] += 1
        obs.check(expected)

        detector.remove_constraints([m.c1])
        expected[m.c1][Reason.removed] += 1
        expected[m.p][Reason.removed] += 1
        obs.check(expected)

        detector.add_constraints([m.c1])
        expected[m.c1][Reason.added] += 1
        expected[m.p][Reason.added] += 1
        obs.check(expected)

        detector.remove_objectives([m.obj])
        expected[m.obj][Reason.removed] += 1
        obs.check(expected)

        detector.add_objectives([m.obj])
        expected[m.obj][Reason.added] += 1
        obs.check(expected)

        m.x.setlb(0)
        detector.update_variables([m.x, m.y])
        expected[m.x][Reason.bounds] += 1
        obs.check(expected)

        m.p.value = 2
        detector.update_parameters([m.p])
        expected[m.p][Reason.value] += 1
        obs.check(expected)

        m.c1.set_value(m.y >= m.x**2)
        detector.update_constraints([m.c1, m.c2])
        expected[m.p][Reason.removed] += 1
        expected[m.c1][Reason.expr] += 1
        obs.check(expected)

        m.obj.expr += m.x
        detector.update_objectives([m.obj])
        expected[m.obj][Reason.expr] += 1
        obs.check(expected)

    def test_mutable_parameters_in_sos(self):
        """
        There is logic in the ModelChangeDetector to handle
        mutable parameters in SOS constraints. However, we cannot
        currently test it because of #3769. For now, we will
        just make sure that an error is raised when attempting to
        use a mutable parameter in an SOS constraint. If #3769 is
        resolved, we will just need to update this test to make
        sure the ModelChangeDetector does the right thing.
        """
        m = pyo.ConcreteModel()
        m.a = pyo.Set(initialize=[1, 2, 3])
        m.x = pyo.Var(m.a)
        m.p = pyo.Param(m.a, mutable=True)
        m.p[1].value = 1
        m.p[2].value = 2
        m.p[3].value = 3

        with self.assertRaisesRegex(
            PyomoException, 'Cannot convert non-constant Pyomo expression .* to bool.*'
        ):
            m.c = pyo.SOSConstraint(var=m.x, sos=1, weights=m.p)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import re
import sys

from pyomo.common.collections import ComponentSet, ComponentMap, Bunch
from pyomo.common.dependencies import attempt_import
from pyomo.common.errors import ApplicationError
from pyomo.common.tempfiles import TempfileManager
from pyomo.common.tee import capture_output
from pyomo.core.expr.numvalue import is_fixed
from pyomo.core.expr.numvalue import value
from pyomo.core.staleflag import StaleFlagManager
from pyomo.repn import generate_standard_repn
from pyomo.solvers.plugins.solvers.direct_solver import DirectSolver
from pyomo.solvers.plugins.solvers.direct_or_persistent_solver import (
    DirectOrPersistentSolver,
)
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.opt.results.results_ import SolverResults
from pyomo.opt.results.solution import Solution, SolutionStatus
from pyomo.opt.results.solver import TerminationCondition, SolverStatus
from pyomo.opt.base import SolverFactory
from pyomo.core.base.suffix import Suffix

logger = logging.getLogger('pyomo.solvers')


class DegreeError(ValueError):
    pass


def _is_numeric(x):
    try:
        float(x)
    except ValueError:
        return False
    return True


def _parse_gurobi_version(gurobipy, avail):
    if not avail:
        return
    GurobiDirect._version = gurobipy.gurobi.version()
    GurobiDirect._name = "Gurobi %s.%s%s" % GurobiDirect._version
    while len(GurobiDirect._version) < 4:
        GurobiDirect._version += (0,)
    GurobiDirect._version = GurobiDirect._version[:4]
    GurobiDirect._version_major = GurobiDirect._version[0]


gurobipy, gurobipy_available = attempt_import(
    'gurobipy',
    # Other forms of exceptions can be thrown by the gurobi python
    # import.  For example, a gurobipy.GurobiError exception is thrown
    # if all tokens for Gurobi are already in use; assuming, of course,
    # the license is a token license.  Unfortunately, you can't import
    # without a license, which means we can't explicitly test for that
    # exception!
    catch_exceptions=(Exception,),
    callback=_parse_gurobi_version,
    defer_import=True,
)


def _set_options(model_or_env, options):
    # Set a parameters from the dictionary 'options' on the given gurobipy
    # model or environment.
    for key, option in options.items():
        # When options come from the pyomo command, all
        # values are string types, so we try to cast
        # them to a numeric value in the event that
        # setting the parameter fails.
        try:
            model_or_env.setParam(key, option)
        except TypeError:
            # we place the exception handling for
            # checking the cast of option to a float in
            # another function so that we can simply
            # call raise here instead of except
            # TypeError as e / raise e, because the
            # latter does not preserve the Gurobi stack
            # trace
            if not _is_numeric(option):
                raise
            model_or_env.setParam(key, float(option))


@SolverFactory.register('gurobi_direct', doc='Direct python interface to Gurobi')
class GurobiDirect(DirectSolver):
    """A direct interface to Gurobi using gurobipy.

    :param manage_env: Set to True if this solver instance should create and
        manage its own Gurobi environment (defaults to False)
    :type manage_env: bool
    :param options: Dictionary of Gurobi parameters to set
    :type options: dict

    If ``manage_env`` is set to True, the ``GurobiDirect`` object creates a local
    Gurobi environment and manage all associated Gurobi resources. Importantly,
    this enables Gurobi licenses to be freed and connections terminated when the
    solver context is exited::

        with SolverFactory('gurobi', solver_io='python', manage_env=True) as opt:
            opt.solve(model)

        # All Gurobi models and environments are freed

    If ``manage_env`` is set to False (the default), the ``GurobiDirect`` object
    uses the global default Gurobi environment::

        with SolverFactory('gurobi', solver_io='python') as opt:
            opt.solve(model)

        # Only models created by `opt` are freed, the global default
        # environment remains active

    ``manage_env=True`` is required when setting license or connection parameters
    programmatically. The ``options`` argument is used to pass parameters to the
    Gurobi environment. For example, to connect to a Gurobi Cluster Manager::

        options = {
            "CSManager": "<url>",
            "CSAPIAccessID": "<access-id>",
            "CSAPISecret": "<api-key>",
        }
        with SolverFactory(
            'gurobi', solver_io='python', manage_env=True, options=options
        ) as opt:
            opt.solve(model)  # Model solved on compute server
        # Compute server connection terminated
    """

    _name = None
    _version = 0
    _version_major = 0
    _default_env_started = False

    def __init__(self, manage_env=False, **kwds):
        if 'type' not in kwds:
            kwds['type'] = 'gurobi_direct'
        super(GurobiDirect, self).__init__(**kwds)
        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._solver_var_to_pyomo_var_map = ComponentMap()
        self._pyomo_con_to_solver_con_map = dict()
        self._solver_con_to_pyomo_con_map = ComponentMap()
        self._needs_updated = True  # flag that indicates if solver_model.update() needs called before getting variable and constraint attributes
        self._callback = None
        self._callback_func = None

        self._python_api_exists = gurobipy_available
        self._range_constraints = set()

        self._max_obj_degree = 2
        self._max_constraint_degree = 2

        # Note: Undefined capabilities default to None
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

        # fix for compatibility with pre-5.0 Gurobi
        #
        # Note: Unfortunately, this will trigger the immediate import
        #    of the gurobipy module
        if gurobipy_available and GurobiDirect._version_major < 5:
            self._max_constraint_degree = 1
            self._capabilities.quadratic_constraint = False

        # remove the instance-level definition of the gurobi version:
        # because the version comes from an imported module, only one
        # version of gurobi is supported (and stored as a class attribute)
        del self._version

        self._manage_env = manage_env
        self._env = None
        self._env_options = None
        self._solver_model = None

    def available(self, exception_flag=True):
        """Returns True if the solver is available.

        :param exception_flag: If True, raise an exception instead of returning
            False if the solver is unavailable (defaults to False)
        :type exception_flag: bool

        In general, ``available()`` does not need to be called by the user, as
        the check is run automatically when solving a model. However it is useful
        for a simple retry loop when using a shared Gurobi license::

            with SolverFactory('gurobi', solver_io='python') as opt:
                while not available(exception_flag=False):
                    time.sleep(1)
                opt.solve(model)

        """
        # First check gurobipy is imported
        if not gurobipy_available:
            if exception_flag:
                gurobipy.log_import_warning(logger=__name__)
                raise ApplicationError(
                    "No Python bindings available for %s solver plugin" % (type(self),)
                )
            return False

        # Ensure environment is started to check for a valid license
        with capture_output(capture_fd=True) as OUT:
            try:
                self._init_env()
                return True
            except gurobipy.GurobiError as e:
                msg = "Could not create Model - gurobi message=%s\n" % (e,)
        if OUT.getvalue():
            msg += "\n" + OUT.getvalue()
        # Didn't return, so environment start failed
        if exception_flag:
            logger.warning(msg)
            raise ApplicationError(
                "Could not create Model for %s solver plugin - gurobi message=%s"
                % (type(self), msg)
            )
        else:
            return False

    def _apply_solver(self):
        StaleFlagManager.mark_all_as_stale()

        # We would like to just set OutputFlag to turn off the console
        # log, but that prevents users from using the Gurobi LogFile
        # option (see #3589 / #3716).  BUT just setting LogToConsole
        # triggers Gurobi to write a message to the console.  We will
        # capture that message here:
        with capture_output(capture_fd=True):
            self._solver_model.setParam('LogToConsole', int(bool(self._tee)))

        if self._keepfiles:
            # Only save log file when the user wants to keep it.
            self._solver_model.setParam('LogFile', self._log_file)
            print("Solver log file: " + self._log_file)

        # Only pass along changed parameters to the model
        if self._env_options:
            new_options = {
                key: option
                for key, option in self.options.items()
                if key not in self._env_options or self._env_options[key] != option
            }
        else:
            new_options = self.options
        _set_options(self._solver_model, new_options)

        if self._version_major >= 5:
            for suffix in self._suffixes:
                if re.match(suffix, "dual"):
                    self._solver_model.setParam(gurobipy.GRB.Param.QCPDual, 1)

        self._solver_model.optimize(self._callback)
        self._needs_updated = False

        if self._keepfiles:
            # Change LogFile to make Gurobi close the original log file.
            # May not work for all Gurobi versions, like ver. 9.5.0.
            self._solver_model.setParam('LogFile', 'default')

        # FIXME: can we get a return code indicating if Gurobi had a significant failure?
        return Bunch(rc=None, log=None)

    def _get_expr_from_pyomo_repn(self, repn, max_degree=2):
        referenced_vars = ComponentSet()

        degree = repn.polynomial_degree()
        if (degree is None) or (degree > max_degree):
            raise DegreeError(
                'GurobiDirect does not support expressions of degree {0}.'.format(
                    degree
                )
            )

        if len(repn.linear_vars) > 0:
            referenced_vars.update(repn.linear_vars)
            new_expr = gurobipy.LinExpr(
                repn.linear_coefs,
                [self._pyomo_var_to_solver_var_map[i] for i in repn.linear_vars],
            )
        else:
            new_expr = 0.0

        for i, v in enumerate(repn.quadratic_vars):
            x, y = v
            new_expr += (
                repn.quadratic_coefs[i]
                * self._pyomo_var_to_solver_var_map[x]
                * self._pyomo_var_to_solver_var_map[y]
            )
            referenced_vars.add(x)
            referenced_vars.add(y)

        new_expr += repn.constant

        return new_expr, referenced_vars, degree

    def _get_expr_from_pyomo_expr(self, expr, max_degree=2):
        if max_degree == 2:
            repn = generate_standard_repn(expr, quadratic=True)
        else:
            repn = generate_standard_repn(expr, quadratic=False)

        try:
            gurobi_expr, referenced_vars, degree = self._get_expr_from_pyomo_repn(
                repn, max_degree
            )
        except DegreeError as e:
            msg = e.args[0]
            msg += '\nexpr: {0}'.format(expr)
            raise DegreeError(msg)

        return gurobi_expr, referenced_vars, degree

    def _gurobi_lb_ub_from_var(self, var):
        if var.is_fixed():
            val = var.value
            return val, val
        if var.has_lb():
            lb = value(var.lb)
        else:
            lb = -gurobipy.GRB.INFINITY
        if var.has_ub():
            ub = value(var.ub)
        else:
            ub = gurobipy.GRB.INFINITY
        return lb, ub

    def _add_var(self, var):
        varname = self._symbol_map.getSymbol(var, self._labeler)
        vtype = self._gurobi_vtype_from_var(var)
        lb, ub = self._gurobi_lb_ub_from_var(var)

        gurobipy_var = self._solver_model.addVar(
            lb=lb, ub=ub, vtype=vtype, name=varname
        )

        self._pyomo_var_to_solver_var_map[var] = gurobipy_var
        self._solver_var_to_pyomo_var_map[gurobipy_var] = var
        self._referenced_variables[var] = 0

        self._needs_updated = True

    def close_global(self):
        """Frees all Gurobi models used by this solver, and frees the global
        default Gurobi environment.

        The default environment is used by all ``GurobiDirect`` solvers started
        with ``manage_env=False`` (the default). To guarantee that all Gurobi
        resources are freed, all instantiated ``GurobiDirect`` solvers must also
        be correctly closed.

        The following example will free all Gurobi resources assuming the user did
        not create any other models (e.g. via another ``GurobiDirect`` object with
        ``manage_env=False``)::

            opt = SolverFactory('gurobi', solver_io='python')
            try:
                opt.solve(model)
            finally:
                opt.close_global()
            # All Gurobi models created by `opt` are freed and the default
            # Gurobi environment is closed
        """
        self.close()
        with capture_output(capture_fd=True):
            gurobipy.disposeDefaultEnv()
        GurobiDirect._default_env_started = False

    def _init_env(self):
        if self._manage_env:
            # Ensure an environment is active for this instance
            if self._env is None:
                assert self._solver_model is None
                env = gurobipy.Env(empty=True)
                _set_options(env, self.options)
                env.start()
                # Successful start (no errors): store the environment
                self._env = env
                self._env_options = dict(self.options)
        else:
            # Ensure the (global) default env is started
            if not GurobiDirect._default_env_started:
                m = gurobipy.Model()
                m.close()
                GurobiDirect._default_env_started = True

    def _create_model(self, model):
        self._init_env()
        if self._solver_model is not None:
            self._solver_model.close()

        self._solver_model = (
            gurobipy.Model(model.name, env=self._env)
            if model.name is not None
            else gurobipy.Model(env=self._env)
        )

    def close(self):
        """Frees local Gurobi resources used by this solver instance.

        All Gurobi models created by the solver are freed. If the solver was
        created with ``manage_env=True``, this method also closes the Gurobi
        environment used by this solver instance. Calling ``.close()`` achieves
        the same result as exiting the solver context (although using context
        managers is preferred where possible)::

            opt = SolverFactory('gurobi', solver_io='python', manage_env=True)
            try:
                opt.solve(model)
            finally:
                opt.close()
            # Gurobi models and environments created by `opt` are freed

        As with the context manager, if ``manage_env=False`` (the default) was
        used, only the Gurobi models created by this solver are freed. The
        default global Gurobi environment will still be active::

            opt = SolverFactory('gurobi', solver_io='python')
            try:
                opt.solve(model)
            finally:
                opt.close()
            # Gurobi models created by `opt` are freed; however the
            # default/global Gurobi environment is still active
        """

        if self._solver_model is not None:
            self._solver_model.close()
            self._solver_model = None
        if self._manage_env:
            if self._env is not None:
                self._env.close()
                self._env = None
                self._env_options = None

    def __exit__(self, t, v, traceback):
        super().__exit__(t, v, traceback)
        self.close()

    def _set_instance(self, model, kwds={}):
        self._range_constraints = set()
        DirectOrPersistentSolver._set_instance(self, model, kwds)
        self._pyomo_con_to_solver_con_map = dict()
        self._solver_con_to_pyomo_con_map = ComponentMap()
        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._solver_var_to_pyomo_var_map = ComponentMap()
        try:
            self._create_model(model)
        except Exception:
            e = sys.exc_info()[1]
            msg = (
                "Unable to create Gurobi model. "
                "Have you installed the Python "
                "bindings for Gurobi?\n\n\t" + "Error message: {0}".format(e)
            )
            raise Exception(msg)

        self._add_block(model)

        for var, n_ref in self._referenced_variables.items():
            if n_ref != 0:
                if var.fixed:
                    if not self._output_fixed_variable_bounds:
                        raise ValueError(
                            "Encountered a fixed variable (%s) inside "
                            "an active objective or constraint "
                            "expression on model %s, which is usually "
                            "indicative of a preprocessing error. Use "
                            "the IO-option 'output_fixed_variable_bounds=True' "
                            "to suppress this error and fix the variable "
                            "by overwriting its bounds in the Gurobi instance."
                            % (var.name, self._pyomo_model.name)
                        )

    def _add_block(self, block):
        DirectOrPersistentSolver._add_block(self, block)

    def _addConstr(self, degree, lhs, sense=None, rhs=None, name=""):
        if degree == 2:
            con = self._solver_model.addQConstr(lhs, sense, rhs, name)
        else:
            con = self._solver_model.addLConstr(lhs, sense, rhs, name)
        return con

    def _add_constraint(self, con):
        if not con.active:
            return None

        if self._skip_trivial_constraints and is_fixed(con.body):
            return None

        conname = self._symbol_map.getSymbol(con, self._labeler)

        if con._linear_canonical_form:
            gurobi_expr, referenced_vars, degree = self._get_expr_from_pyomo_repn(
                con.canonical_form(), self._max_constraint_degree
            )
        else:
            gurobi_expr, referenced_vars, degree = self._get_expr_from_pyomo_expr(
                con.body, self._max_constraint_degree
            )

        if con.has_lb():
            if not is_fixed(con.lower):
                raise ValueError(
                    "Lower bound of constraint {0} is not constant.".format(con)
                )
        if con.has_ub():
            if not is_fixed(con.upper):
                raise ValueError(
                    "Upper bound of constraint {0} is not constant.".format(con)
                )

        if con.equality:
            gurobipy_con = self._addConstr(
                degree=degree,
                lhs=gurobi_expr,
                sense=gurobipy.GRB.EQUAL,
                rhs=value(con.lower),
                name=conname,
            )
        elif con.has_lb() and con.has_ub():
            gurobipy_con = self._solver_model.addRange(
                gurobi_expr, value(con.lower), value(con.upper), name=conname
            )
            self._range_constraints.add(con)
        elif con.has_lb():
            gurobipy_con = self._addConstr(
                degree=degree,
                lhs=gurobi_expr,
                sense=gurobipy.GRB.GREATER_EQUAL,
                rhs=value(con.lower),
                name=conname,
            )
        elif con.has_ub():
            gurobipy_con = self._addConstr(
                degree=degree,
                lhs=gurobi_expr,
                sense=gurobipy.GRB.LESS_EQUAL,
                rhs=value(con.upper),
                name=conname,
            )
        else:
            raise ValueError(
                "Constraint does not have a lower "
                "or an upper bound: {0} \n".format(con)
            )

        for var in referenced_vars:
            self._referenced_variables[var] += 1
        self._vars_referenced_by_con[con] = referenced_vars
        self._pyomo_con_to_solver_con_map[con] = gurobipy_con
        self._solver_con_to_pyomo_con_map[gurobipy_con] = con

        self._needs_updated = True

    def _add_sos_constraint(self, con):
        if not con.active:
            return None

        conname = self._symbol_map.getSymbol(con, self._labeler)
        level = con.level
        if level == 1:
            sos_type = gurobipy.GRB.SOS_TYPE1
        elif level == 2:
            sos_type = gurobipy.GRB.SOS_TYPE2
        else:
            raise ValueError(
                "Solver does not support SOS level {0} constraints".format(level)
            )

        gurobi_vars = []
        weights = []

        self._vars_referenced_by_con[con] = ComponentSet()

        if hasattr(con, 'get_items'):
            # aml sos constraint
            sos_items = list(con.get_items())
        else:
            # kernel sos constraint
            sos_items = list(con.items())

        for v, w in sos_items:
            self._vars_referenced_by_con[con].add(v)
            gurobi_vars.append(self._pyomo_var_to_solver_var_map[v])
            self._referenced_variables[v] += 1
            weights.append(w)

        gurobipy_con = self._solver_model.addSOS(sos_type, gurobi_vars, weights)
        self._pyomo_con_to_solver_con_map[con] = gurobipy_con
        self._solver_con_to_pyomo_con_map[gurobipy_con] = con

        self._needs_updated = True

    def _gurobi_vtype_from_var(self, var):
        """
        This function takes a pyomo variable and returns the appropriate gurobi variable type
        :param var: pyomo.core.base.var.Var
        :return: gurobipy.GRB.CONTINUOUS or gurobipy.GRB.BINARY or gurobipy.GRB.INTEGER
        """
        if var.is_binary():
            vtype = gurobipy.GRB.BINARY
        elif var.is_integer():
            vtype = gurobipy.GRB.INTEGER
        elif var.is_continuous():
            vtype = gurobipy.GRB.CONTINUOUS
        else:
            raise ValueError(
                'Variable domain type is not recognized for {0}'.format(var.domain)
            )
        return vtype

    def _set_objective(self, obj):
        if self._objective is not None:
            for var in self._vars_referenced_by_obj:
                self._referenced_variables[var] -= 1
            self._vars_referenced_by_obj = ComponentSet()
            self._objective = None

        if obj.active is False:
            raise ValueError('Cannot add inactive objective to solver.')

        if obj.sense == minimize:
            sense = gurobipy.GRB.MINIMIZE
        elif obj.sense == maximize:
            sense = gurobipy.GRB.MAXIMIZE
        else:
            raise ValueError('Objective sense is not recognized: {0}'.format(obj.sense))

        gurobi_expr, referenced_vars, degree = self._get_expr_from_pyomo_expr(
            obj.expr, self._max_obj_degree
        )

        for var in referenced_vars:
            self._referenced_variables[var] += 1

        self._solver_model.setObjective(gurobi_expr, sense=sense)
        self._objective = obj
        self._vars_referenced_by_obj = referenced_vars

        self._needs_updated = True

    def _postsolve(self):
        # the only suffixes that we extract from GUROBI are
        # constraint duals, constraint slacks, and variable
        # reduced-costs. scan through the solver suffix list
        # and throw an exception if the user has specified
        # any others.
        extract_duals = False
        extract_slacks = False
        extract_reduced_costs = False
        for suffix in self._suffixes:
            flag = False
            if re.match(suffix, "dual"):
                extract_duals = True
                flag = True
            if re.match(suffix, "slack"):
                extract_slacks = True
                flag = True
            if re.match(suffix, "rc"):
                extract_reduced_costs = True
                flag = True
            if not flag:
                raise RuntimeError(
                    "***The gurobi_direct solver plugin cannot extract solution suffix="
                    + suffix
                )

        gprob = self._solver_model
        grb = gurobipy.GRB
        status = gprob.Status

        if gprob.getAttr(gurobipy.GRB.Attr.IsMIP):
            if extract_reduced_costs:
                logger.warning("Cannot get reduced costs for MIP.")
            if extract_duals:
                logger.warning("Cannot get duals for MIP.")
            extract_reduced_costs = False
            extract_duals = False

        self.results = SolverResults()
        soln = Solution()

        self.results.solver.name = GurobiDirect._name
        self.results.solver.wallclock_time = gprob.Runtime

        if status == grb.LOADED:  # problem is loaded, but no solution
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "Model is loaded, but no solution information is available."
            )
            self.results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.unknown
        elif status == grb.OPTIMAL:  # optimal
            self.results.solver.status = SolverStatus.ok
            self.results.solver.termination_message = (
                "Model was solved to optimality (subject to tolerances), "
                "and an optimal solution is available."
            )
            self.results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif status == grb.INFEASIBLE:
            self.results.solver.status = SolverStatus.warning
            self.results.solver.termination_message = (
                "Model was proven to be infeasible"
            )
            self.results.solver.termination_condition = TerminationCondition.infeasible
            soln.status = SolutionStatus.infeasible
        elif status == grb.INF_OR_UNBD:
            self.results.solver.status = SolverStatus.warning
            self.results.solver.termination_message = (
                "Problem proven to be infeasible or unbounded."
            )
            self.results.solver.termination_condition = (
                TerminationCondition.infeasibleOrUnbounded
            )
            soln.status = SolutionStatus.unsure
        elif status == grb.UNBOUNDED:
            self.results.solver.status = SolverStatus.warning
            self.results.solver.termination_message = (
                "Model was proven to be unbounded."
            )
            self.results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        elif status == grb.CUTOFF:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "Optimal objective for model was proven to be worse than the "
                "value specified in the Cutoff parameter. No solution "
                "information is available."
            )
            self.results.solver.termination_condition = (
                TerminationCondition.minFunctionValue
            )
            soln.status = SolutionStatus.unknown
        elif status == grb.ITERATION_LIMIT:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "Optimization terminated because the total number of simplex "
                "iterations performed exceeded the value specified in the "
                "IterationLimit parameter."
            )
            self.results.solver.termination_condition = (
                TerminationCondition.maxIterations
            )
            soln.status = SolutionStatus.stoppedByLimit
        elif status == grb.NODE_LIMIT:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "Optimization terminated because the total number of "
                "branch-and-cut nodes explored exceeded the value specified "
                "in the NodeLimit parameter"
            )
            self.results.solver.termination_condition = (
                TerminationCondition.maxEvaluations
            )
            soln.status = SolutionStatus.stoppedByLimit
        elif status == grb.TIME_LIMIT:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "Optimization terminated because the time expended exceeded "
                "the value specified in the TimeLimit parameter."
            )
            self.results.solver.termination_condition = (
                TerminationCondition.maxTimeLimit
            )
            soln.status = SolutionStatus.stoppedByLimit
        elif status == grb.SOLUTION_LIMIT:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "Optimization terminated because the number of solutions found "
                "reached the value specified in the SolutionLimit parameter."
            )
            self.results.solver.termination_condition = TerminationCondition.unknown
            soln.status = SolutionStatus.stoppedByLimit
        elif status == grb.INTERRUPTED:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "Optimization was terminated by the user."
            )
            self.results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error
        elif status == grb.NUMERIC:
            self.results.solver.status = SolverStatus.error
            self.results.solver.termination_message = (
                "Optimization was terminated due to unrecoverable numerical "
                "difficulties."
            )
            self.results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error
        elif status == grb.SUBOPTIMAL:
            self.results.solver.status = SolverStatus.warning
            self.results.solver.termination_message = (
                "Unable to satisfy optimality tolerances; a sub-optimal "
                "solution is available."
            )
            self.results.solver.termination_condition = TerminationCondition.other
            soln.status = SolutionStatus.feasible
        # note that USER_OBJ_LIMIT was added in Gurobi 7.0, so it may not be present
        elif (status is not None) and (status == getattr(grb, 'USER_OBJ_LIMIT', None)):
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_message = (
                "User specified an objective limit "
                "(a bound on either the best objective "
                "or the best bound), and that limit has "
                "been reached. Solution is available."
            )
            self.results.solver.termination_condition = TerminationCondition.other
            soln.status = SolutionStatus.stoppedByLimit
        else:
            self.results.solver.status = SolverStatus.error
            self.results.solver.termination_message = (
                "Unhandled Gurobi solve status (" + str(status) + ")"
            )
            self.results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error

        self.results.problem.name = gprob.ModelName

        if gprob.ModelSense == 1:
            self.results.problem.sense = minimize
        elif gprob.ModelSense == -1:
            self.results.problem.sense = maximize
        else:
            raise RuntimeError(
                'Unrecognized gurobi objective sense: {0}'.format(gprob.ModelSense)
            )

        self.results.problem.upper_bound = None
        self.results.problem.lower_bound = None
        if (gprob.NumBinVars + gprob.NumIntVars) == 0:
            try:
                self.results.problem.upper_bound = gprob.ObjVal
                self.results.problem.lower_bound = gprob.ObjVal
            except (gurobipy.GurobiError, AttributeError):
                pass
        elif gprob.ModelSense == 1:  # minimizing
            try:
                self.results.problem.upper_bound = gprob.ObjVal
            except (gurobipy.GurobiError, AttributeError):
                pass
            try:
                self.results.problem.lower_bound = gprob.ObjBound
            except (gurobipy.GurobiError, AttributeError):
                pass
        elif gprob.ModelSense == -1:  # maximizing
            try:
                self.results.problem.upper_bound = gprob.ObjBound
            except (gurobipy.GurobiError, AttributeError):
                pass
            try:
                self.results.problem.lower_bound = gprob.ObjVal
            except (gurobipy.GurobiError, AttributeError):
                pass
        else:
            raise RuntimeError(
                'Unrecognized gurobi objective sense: {0}'.format(gprob.ModelSense)
            )

        try:
            soln.gap = (
                self.results.problem.upper_bound - self.results.problem.lower_bound
            )
        except TypeError:
            soln.gap = None

        self.results.problem.number_of_constraints = (
            gprob.NumConstrs + gprob.NumQConstrs + gprob.NumSOS
        )
        self.results.problem.number_of_nonzeros = gprob.NumNZs
        self.results.problem.number_of_variables = gprob.NumVars
        self.results.problem.number_of_binary_variables = gprob.NumBinVars
        self.results.problem.number_of_integer_variables = gprob.NumIntVars
        self.results.problem.number_of_continuous_variables = (
            gprob.NumVars - gprob.NumIntVars - gprob.NumBinVars
        )
        self.results.problem.number_of_objectives = 1
        self.results.problem.number_of_solutions = gprob.SolCount

        # if a solve was stopped by a limit, we still need to check to
        # see if there is a solution available - this may not always
        # be the case, both in LP and MIP contexts.
        if self._save_results:
            """
            This code in this if statement is only needed for backwards compatibility. It is more efficient to set
            _save_results to False and use load_vars, load_duals, etc.
            """
            if gprob.SolCount > 0:
                soln_variables = soln.variable
                soln_constraints = soln.constraint

                gurobi_vars = self._solver_model.getVars()
                gurobi_vars = list(
                    set(gurobi_vars).intersection(
                        set(self._pyomo_var_to_solver_var_map.values())
                    )
                )
                var_vals = self._solver_model.getAttr("X", gurobi_vars)
                names = self._solver_model.getAttr("VarName", gurobi_vars)
                for gurobi_var, val, name in zip(gurobi_vars, var_vals, names):
                    pyomo_var = self._solver_var_to_pyomo_var_map[gurobi_var]
                    if self._referenced_variables[pyomo_var] > 0:
                        soln_variables[name] = {"Value": val}

                if extract_reduced_costs:
                    vals = self._solver_model.getAttr("Rc", gurobi_vars)
                    for gurobi_var, val, name in zip(gurobi_vars, vals, names):
                        pyomo_var = self._solver_var_to_pyomo_var_map[gurobi_var]
                        if self._referenced_variables[pyomo_var] > 0:
                            soln_variables[name]["Rc"] = val

                if extract_duals or extract_slacks:
                    gurobi_cons = self._solver_model.getConstrs()
                    con_names = self._solver_model.getAttr("ConstrName", gurobi_cons)
                    for name in con_names:
                        soln_constraints[name] = {}
                    if self._version_major >= 5:
                        gurobi_q_cons = self._solver_model.getQConstrs()
                        q_con_names = self._solver_model.getAttr(
                            "QCName", gurobi_q_cons
                        )
                        for name in q_con_names:
                            soln_constraints[name] = {}

                if extract_duals:
                    vals = self._solver_model.getAttr("Pi", gurobi_cons)
                    for val, name in zip(vals, con_names):
                        soln_constraints[name]["Dual"] = val
                    if self._version_major >= 5:
                        q_vals = self._solver_model.getAttr("QCPi", gurobi_q_cons)
                        for val, name in zip(q_vals, q_con_names):
                            soln_constraints[name]["Dual"] = val

                if extract_slacks:
                    gurobi_range_con_vars = set(self._solver_model.getVars()) - set(
                        self._pyomo_var_to_solver_var_map.values()
                    )
                    vals = self._solver_model.getAttr("Slack", gurobi_cons)
                    for gurobi_con, val, name in zip(gurobi_cons, vals, con_names):
                        pyomo_con = self._solver_con_to_pyomo_con_map[gurobi_con]
                        if pyomo_con in self._range_constraints:
                            lin_expr = self._solver_model.getRow(gurobi_con)
                            for i in reversed(range(lin_expr.size())):
                                v = lin_expr.getVar(i)
                                if v in gurobi_range_con_vars:
                                    Us_ = v.X
                                    Ls_ = v.UB - v.X
                                    if Us_ > Ls_:
                                        soln_constraints[name]["Slack"] = Us_
                                    else:
                                        soln_constraints[name]["Slack"] = -Ls_
                                    break
                        else:
                            soln_constraints[name]["Slack"] = val
                    if self._version_major >= 5:
                        q_vals = self._solver_model.getAttr("QCSlack", gurobi_q_cons)
                        for val, name in zip(q_vals, q_con_names):
                            soln_constraints[name]["Slack"] = val
        elif self._load_solutions:
            if gprob.SolCount > 0:
                self.load_vars()

                if extract_reduced_costs:
                    self._load_rc()

                if extract_duals:
                    self._load_duals()

                if extract_slacks:
                    self._load_slacks()

        self.results.solution.insert(soln)

        # finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin.
        TempfileManager.pop(remove=not self._keepfiles)

        return DirectOrPersistentSolver._postsolve(self)

    def warm_start_capable(self):
        return True

    def _warm_start(self):
        for pyomo_var, gurobipy_var in self._pyomo_var_to_solver_var_map.items():
            if pyomo_var.value is not None:
                gurobipy_var.setAttr(gurobipy.GRB.Attr.Start, value(pyomo_var))
        self._needs_updated = True

    def _load_vars(self, vars_to_load=None):
        var_map = self._pyomo_var_to_solver_var_map
        ref_vars = self._referenced_variables
        if vars_to_load is None:
            vars_to_load = var_map.keys()

        gurobi_vars_to_load = [var_map[pyomo_var] for pyomo_var in vars_to_load]
        vals = self._solver_model.getAttr("X", gurobi_vars_to_load)

        for var, val in zip(vars_to_load, vals):
            if ref_vars[var] > 0:
                var.set_value(val, skip_validation=True)

    def _load_rc(self, vars_to_load=None):
        if not hasattr(self._pyomo_model, 'rc'):
            self._pyomo_model.rc = Suffix(direction=Suffix.IMPORT)
        var_map = self._pyomo_var_to_solver_var_map
        ref_vars = self._referenced_variables
        rc = self._pyomo_model.rc
        if vars_to_load is None:
            vars_to_load = var_map.keys()

        gurobi_vars_to_load = [var_map[pyomo_var] for pyomo_var in vars_to_load]
        vals = self._solver_model.getAttr("Rc", gurobi_vars_to_load)

        for var, val in zip(vars_to_load, vals):
            if ref_vars[var] > 0:
                rc[var] = val

    def _load_duals(self, cons_to_load=None):
        if not hasattr(self._pyomo_model, 'dual'):
            self._pyomo_model.dual = Suffix(direction=Suffix.IMPORT)
        con_map = self._pyomo_con_to_solver_con_map
        reverse_con_map = self._solver_con_to_pyomo_con_map
        dual = self._pyomo_model.dual

        if cons_to_load is None:
            linear_cons_to_load = self._solver_model.getConstrs()
            if self._version_major >= 5:
                quadratic_cons_to_load = self._solver_model.getQConstrs()
        else:
            gurobi_cons_to_load = set(
                [con_map[pyomo_con] for pyomo_con in cons_to_load]
            )
            linear_cons_to_load = gurobi_cons_to_load.intersection(
                set(self._solver_model.getConstrs())
            )
            if self._version_major >= 5:
                quadratic_cons_to_load = gurobi_cons_to_load.intersection(
                    set(self._solver_model.getQConstrs())
                )
        linear_vals = self._solver_model.getAttr("Pi", linear_cons_to_load)
        if self._version_major >= 5:
            quadratic_vals = self._solver_model.getAttr("QCPi", quadratic_cons_to_load)

        for gurobi_con, val in zip(linear_cons_to_load, linear_vals):
            pyomo_con = reverse_con_map[gurobi_con]
            dual[pyomo_con] = val
        if self._version_major >= 5:
            for gurobi_con, val in zip(quadratic_cons_to_load, quadratic_vals):
                pyomo_con = reverse_con_map[gurobi_con]
                dual[pyomo_con] = val

    def _load_slacks(self, cons_to_load=None):
        if not hasattr(self._pyomo_model, 'slack'):
            self._pyomo_model.slack = Suffix(direction=Suffix.IMPORT)
        con_map = self._pyomo_con_to_solver_con_map
        reverse_con_map = self._solver_con_to_pyomo_con_map
        slack = self._pyomo_model.slack

        gurobi_range_con_vars = set(self._solver_model.getVars()) - set(
            self._pyomo_var_to_solver_var_map.values()
        )

        if cons_to_load is None:
            linear_cons_to_load = self._solver_model.getConstrs()
            if self._version_major >= 5:
                quadratic_cons_to_load = self._solver_model.getQConstrs()
        else:
            gurobi_cons_to_load = set(
                [con_map[pyomo_con] for pyomo_con in cons_to_load]
            )
            linear_cons_to_load = gurobi_cons_to_load.intersection(
                set(self._solver_model.getConstrs())
            )
            if self._version_major >= 5:
                quadratic_cons_to_load = gurobi_cons_to_load.intersection(
                    set(self._solver_model.getQConstrs())
                )
        linear_vals = self._solver_model.getAttr("Slack", linear_cons_to_load)
        if self._version_major >= 5:
            quadratic_vals = self._solver_model.getAttr(
                "QCSlack", quadratic_cons_to_load
            )

        for gurobi_con, val in zip(linear_cons_to_load, linear_vals):
            pyomo_con = reverse_con_map[gurobi_con]
            if pyomo_con in self._range_constraints:
                lin_expr = self._solver_model.getRow(gurobi_con)
                for i in reversed(range(lin_expr.size())):
                    v = lin_expr.getVar(i)
                    if v in gurobi_range_con_vars:
                        Us_ = v.X
                        Ls_ = v.UB - v.X
                        if Us_ > Ls_:
                            slack[pyomo_con] = Us_
                        else:
                            slack[pyomo_con] = -Ls_
                        break
            else:
                slack[pyomo_con] = val
        if self._version_major >= 5:
            for gurobi_con, val in zip(quadratic_cons_to_load, quadratic_vals):
                pyomo_con = reverse_con_map[gurobi_con]
                slack[pyomo_con] = val

    def load_duals(self, cons_to_load=None):
        """
        Load the duals into the 'dual' suffix. The 'dual' suffix must live on the parent model.

        Parameters
        ----------
        cons_to_load: list of Constraint
        """
        self._load_duals(cons_to_load)

    def load_rc(self, vars_to_load):
        """
        Load the reduced costs into the 'rc' suffix. The 'rc' suffix must live on the parent model.

        Parameters
        ----------
        vars_to_load: list of Var
        """
        self._load_rc(vars_to_load)

    def load_slacks(self, cons_to_load=None):
        """
        Load the values of the slack variables into the 'slack' suffix. The 'slack' suffix must live on the parent
        model.

        Parameters
        ----------
        cons_to_load: list of Constraint
        """
        self._load_slacks(cons_to_load)

    def _update(self):
        self._solver_model.update()
        self._needs_updated = False
